/*
 * File access convenience functions and types.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <sys/stat.h>

#include <l4/re/c/rm.h>
#include <l4/sys/types.h>

#include <systypes/base.h>
#include <systypes/user.h>



/* C compatibility types (defined in the implementation). */

struct file_notifier;

typedef struct file_notifier file_notifier_t;



EXTERN_C_BEGIN

/* File access abstraction. */

typedef struct
{
  /* File object reference. */

  l4_cap_idx_t ref;

  /* Mapped memory accessing a file region. */

  char *memory;

  /* File region parameters. */

  offset_t start_pos, end_pos;  /* start and end positions of region */
  offset_t data_end;            /* amount/extent of data in the region */
  offset_t data_current;        /* client access offset */

  /* Total size of file. */

  offset_t size;

  /* Support for arbitrary memory mapping and explicit object size. */

  object_flags_t object_flags;

  /* Blocking accesses. */

  notify_flags_t can_block;

  /* Saved notifications. */

  notify_flags_t notifications;

  /* Flags indicated when opening the file. */

  flags_t flags;

  /* Any failure or error condition on the file. */

  long error;

} file_t;



/* Filesystem operations. */

long file_open_for_user(user_t user, l4_cap_idx_t server, l4_cap_idx_t *opener);

/* File operations. */

void file_close(file_t *file);
long file_mkdir(const char *filename, mode_t mode, l4_cap_idx_t server);
long file_open(file_t *file, const char *filename, flags_t flags, l4_cap_idx_t server);
long file_remove(const char *filename, l4_cap_idx_t server);
long file_reopen(file_t *file, file_t *new_file, flags_t flags);
long file_rename(const char *source, const char *target, l4_cap_idx_t server);
long file_stat(const char *filename, struct stat *st, l4_cap_idx_t server);

/* File lifecycle operations. */

long file_context(file_t *file, l4_cap_idx_t server);
long file_context_mkdir(mode_t mode, file_t *context);
long file_context_open(file_t *file, flags_t flags, file_t *context);
long file_context_remove(file_t *context);
long file_context_rename(file_t *context);
long file_context_stat(struct stat *st, file_t *context);
void file_init(file_t *file);

/* File and region operations. */

long file_flush(file_t *file);
long file_mmap(file_t *file, offset_t position, offset_t length,
               offset_t start_visible, offset_t end_visible,
               l4re_rm_flags_t region_flags);
long file_mmap_only(file_t *file, offset_t position, offset_t length,
                    offset_t start_visible, offset_t end_visible);
flags_t file_opening_flags(l4re_rm_flags_t rm_flags);
l4re_rm_flags_t file_region_flags(flags_t flags);
long file_resize(file_t *file, offset_t size);

/* File and region properties. */

offset_t file_populated_span(file_t *file);
offset_t file_span(file_t *file);

/* Convenience functions. */

char *file_string_get(file_t *file, offset_t offset);
int file_string_set(file_t *file, const char *data, offset_t offset, offset_t *written);

/* Client data functions. */

offset_t file_data_available(file_t *file);
char *file_data_current(file_t *file);
offset_t file_data_current_position(file_t *file);
offset_t file_data_end_position(file_t *file);
offset_t file_data_space(file_t *file);

/* Client data transfer functions. */

void file_data_read(file_t *file, char *buf, offset_t to_transfer);
void file_data_write(file_t *file, char *buf, offset_t to_transfer);

/* Notification functions. */

void file_notify_close(file_notifier_t *notifier);
file_notifier_t *file_notify_local(void);
file_notifier_t *file_notify_task(void);

long file_notify_subscribe(file_t *file, notify_flags_t flags, file_notifier_t *notifier);
long file_notify_unsubscribe(file_t *file, file_notifier_t *notifier);
long file_notify_wait_file(file_t *file, file_notifier_t *notifier);
long file_notify_wait_files(file_t **file, file_notifier_t *notifier);



/* Pipe operations. */

long pipe_open(offset_t size, file_t *reader, file_t *writer, l4_cap_idx_t server);

/* Pipe region operations. */

long pipe_current(file_t *pipe);
long pipe_next(file_t *pipe);
long pipe_written(file_t *pipe, offset_t size);



/* Directory operations. */

long directory_opendir(file_t *file, file_t *reader);



EXTERN_C_END

// vim: tabstop=2 expandtab shiftwidth=2
